#ifndef H_FWMLAYOUT
#define H_FWMLAYOUT

#include "utils.h"

#include <X11/Xlib.h>

#define BWID 1

#define SNAPWID 10

#define PRTCL_DELETE  1
#define PRTCL_FOCUS   2

#define MY_HINT_URGNT 1
#define MY_HINT_NOFCS 2
#define MY_HINT_FIXED 8
#define MY_HINT_NOBRD 16  //no borders (for dockers)
#define MY_HINT_FLOAT 32  //dialogs, transients
#define MY_HINT_DIALG 64  //dialog, shold be centered to the main app
#define MY_HINT_TYPES 128 //type has been set

#define _NET_WM_STATE_REMOVE     0
#define _NET_WM_STATE_ADD        1
#define _NET_WM_STATE_TOGGLE     2

enum { StickNormal, StickStick, StickSlippery, StickAvoid };
enum { StckDesktop, StckBottom, StckNormal, StckTop, StckDock };
enum { AddlvlSticky, AddlvlEmpty, AddlvlFree, AddlvlOther };
enum { FloatFloat, FloatMax, FloatFull };

enum { CycleAll, CycleTiles, CycleTileFwd, CycleTileBwd };

typedef struct Monitor_ Monitor;
typedef struct Layer_ Layer;
typedef struct Tile_ Tile;
typedef struct Client_ Client;
typedef struct Dsktp_ Dsktp;
typedef struct Rule_ Rule;

#define RULE_INTNONE 0xffff

struct Rule_ {
  char* name;
  char* role;
  char* classN;
  char* classC;
  int x,y,w,h;
  float xr,yr,wr,hr;

  Tile* target;
  int visible_only;
  int no_steal;

  int float_state;
  int stack_level;

  Rule* next;
};

struct Dsktp_ {
  char name[256];
  char layout[64];
  int id;
  Dsktp* prev;
  Dsktp* next;
  Monitor* mon;
};

struct Monitor_ {
  int mw, mh; //screen size

  Client* focus;
  Dsktp* desktop;

  int clientcount; //linked list that handles entire stack
  Client* firstclient;
  Client* lastclient;

  Client* firstfocus;
  Client* lastfocus;

  int layercount;
  Layer* firstlayer; //first is unerasable
  Layer* lastlayer;

  int desktopcount;
  Dsktp* firstdesktop;
  Dsktp* lastdesktop;

  Layer* sfirstlayer;
  Layer* slastlayer;

  Rule* rulz;
};

struct Layer_ {
  char name[256];

  int visible;
  int stacktype;//normal,bottom,top
  int focusorhide;

  int tilecount;//each layer has linked list of tiles
  Tile* firsttile;
  Tile* lasttile;

  int clientcount;

  Layer* next;//double-linked list
  Layer* prev;

  Layer* snext;//double-linked list
  Layer* sprev;

  Monitor* mon;
};

struct Tile_ {
  char name[256];

  int x,y,w,h;
  int xa,ya,wa,ha;
  float xr,yr,wr,hr;

  int force_maximize; //force the clients to be maximized
  int max_clients; //no more than that (except transients)
  int sticky; //collects the next window (and then clears this tag)
  int skip_focus;

  int clientcount;
  Client* firstclient;
  Client* lastclient;

  Tile* next;
  Tile* prev;

  Layer* layer;
};

struct Client_ {
  char name[256];
  char role[256];
  char classN[256];
  char classC[256];

  //here come other client props
  //geometry
  int x,y,w,h;
  //actual geometry
  int xx,yy,ww,hh;
  //limits
  int basew,baseh;
  int incw,inch;
  int maxw,maxh;
  int minw,minh;
  float mina,maxa;

  int bw;

  int visible;//this tells you if you have to remap it or not
  int float_state;//floating, maximized, full_screen
  int stacktype;//normal,bottom,top
  int hints;//urgent,fixed
  int protocols;//delete, take focus

  int holdmove;
  long recentunmapserial;

  Client* next;//list of all clients
  Client* prev;//list of all clients

  Client* snext;//stack of all clients (first = top most)
  Client* sprev;//stack of all clients (first = top most)

  Client* fnext;
  Client* fprev;

  Client* owner;
  Client* children;

  Client* chld_next;

  Tile* tile;
  Layer* layer;
  Monitor* mon;

  Window win;
};

//desktop list
int desktop_remove(Dsktp* dsk);
int desktop_add(Monitor* mon, Dsktp* dsk);

//layer list
int layer_remove(Layer* lay);
int layer_add(Monitor* mon, Layer* lay);

//tile list
int tile_remove(Tile* t);
int tile_add(Layer* lay, Tile* t);

//client list
int client_remove(Client* c);
int client_add(Monitor* mon, Client* c);


//layer stacking
int layer_stack_remove(Layer* lay);
int layer_stack_add_front(Monitor* mon, Layer* lay);
int layer_stack_add_back(Monitor* mon, Layer* lay);
int layer_stack_to_front(Layer* lay);
int layer_stack_to_back(Layer* lay);


//client stacking
int client_stack_remove(Client* c);
int client_stack_add_front(Tile* t, Client* c);
int client_stack_add_back(Tile* t, Client* c);
int client_stack_add_to_layer(Layer* lay, Client* c, int level);//add somewhere to layer (return 1 if unsuccessful)
int client_stack_apply_rules(Monitor* mon, Client* c);
int client_stack_add_to_monitor(Monitor* mon, Client* c);
int client_post_apply_rules(Monitor* mon, Client* c);
Rule* client_find_rule(Monitor* mon, Client* c);
int client_move_to_tile(Tile* t, Client* c);

//client focusing

int client_focus_remove(Client* c);
int client_focus_add_front(Client* c);
int client_focus_add_second(Client* c);
int client_focus_add_back(Client* c);
int client_focus_to_front(Client* c);
int client_focus_to_back(Client* c);


//transients
int client_addtransient(Client* parent, Client* trans);
int client_remtransient(Client* trans);

//calculation
int tile_recalc(Tile* t);

//keep inside and snap
void client_keepintile(Tile* t, int* x, int* y);
//internal: apply hints (affects xywh), the unrestricted window size
int client_movecorner(Client* c, int leftright, int topbottom);
int client_rehint(Client* c, int snap);
//internal: fit to tile and apply maximize/fullscreen (affects xx...), calls rehint
int client_reshape(Client* c, int snap);

int client_move(Client* c, int x, int y, int x0, int y0, Client* oldc);
int client_resize(Client* c, int x, int y, int x0, int y0, Client* oldc);

//rules
Rule* add_rule(Monitor* mon, char* name, char* role, char* classN, char* classC, Tile* t);
void remove_rule(Monitor* mon, Rule* r);

void destroy_rule(Rule* r);

//creation
Client* new_client(Window w);
Tile* new_tile(const char* name, int x, int y, int w, int h);
Tile* new_tile_full(const char* name, int x, int y, int w, int h, float xr, float yr, float wr, float hr);
Layer* new_layer(const char* name);
Monitor* new_monitor(int w, int h);
Dsktp* new_desktop(const char* name, const char* layout);


//lookup

int tile_is_free(Tile* t);

Tile* tile_cycle(Tile* t);
Layer* layer_cycle(Layer* l);
Client* client_cycle(Client* c, int mode);
Client* client_recycle(Client* c, int mode);

Client* win_to_client(Monitor* mon, Window w);
Client* str_to_client(Monitor* mon, const char* name);
Layer* str_to_layer(Monitor* mon, const char* name);
Tile* str_to_tile(Monitor* mon, const char* name);
Tile* str_to_ltile(Layer* l, const char* name);
Tile* id_to_ltile(Layer* l, int id);
Layer* id_to_layer(Monitor* mon, int id);
int layer_to_id(Monitor* mon, Layer* lay);
Dsktp* id_to_desktop(Monitor* mon, int id);
Dsktp* str_to_desktop(Monitor* mon, const char* name);

void client_print(Client* c);
//void desktops_print(Monitor* m);

typedef struct {
  int x1,y1,x2,y2;
} Rect;


//rectangle tools

#define MIN_RECT 16

int rect_isok(Rect* r);
int rect_intersect(Rect* r1, Rect* r2, Rect* r);
int rect_generate(Client* c, Rect* ptr);
int rect_apply(Rect* ptr, Rect* target, Rect* out, int* outcnt, int limit);
int rect_findbest(Client* c, Tile* t);


#endif
